use core::str;
use std::{
    io::{self, Read, Write},
    net::{Shutdown, TcpListener, TcpStream},
    str::Utf8Error,
    sync::Arc,
    thread,
};

use dashmap::{DashMap, mapref::one::Ref};
use httparse::{Request, Response};
use thiserror::Error;

use crate::{config, constants, utils::headers::HeadersExt};

pub fn start_listener_loop(config_arc: &Arc<config::Schema>, listener: &TcpListener) -> ! {
    for stream in listener.incoming() {
        match stream {
            Ok(mut client) => {
                let config_arc = config_arc.clone();
                thread::spawn(|| {
                    if let Err(err) = handle_client(&mut client, config_arc.as_ref()) {
                        eprintln!("err: invalid req head ({err}), closing...");
                        _ = client.shutdown(Shutdown::Both);
                    }
                    drop(client);
                    drop(config_arc);
                });
            }
            Err(err) => eprintln!("error with an incoming listener {err:#?}"),
        }
    }

    unreachable!("listener had to be killed unexpectedly");
}

#[derive(Error, Debug)]
pub enum ClientError {
    #[error("error reading request head: {0}")]
    HeadReadError(io::Error),
    #[error("failed to find \"host\" header")]
    HostHeaderNotFound,
    #[error("requested host {0:?}, but it's not registered")]
    HostNotRegistered(String),
    #[error("failed to connect to backend: {0}")]
    BackendConnectFail(io::Error),
    #[error("io error exchanging client and backend: {0}")]
    ExchangeIoError(io::Error),
    #[error("io error capturing {0} buffer: {0}")]
    BufferError(&'static str, io::Error),
    #[error("io error writing auth headers: {0}")]
    AuthHeadersIoError(io::Error),
    #[error("failed to parse response head: {0}")]
    ParseResponseHead(httparse::Error),
    #[error("attempted to parse an unfinished response")]
    UnfinishedResponse,
    #[error("error with text encoding: {0}")]
    TextEncodingError(Utf8Error),
    #[error("error building html csrf finder selector: {0}")]
    SelectorBuildError(String),
    #[error("couldn't find csrf token in response")]
    CsrfNotFound,
    #[error("invalid header {0:?}")]
    InvalidHeader(&'static str),
}

// fn(req, (adrr, etc)) -> Action
// Action {
//  CaptureBody?
//  CaptureResponse?
// }
//
// cases:
// * GET login -> get response to access csrf token
// * POST login -> get body to handle authentication
//
// auth headers are just added depending on the ID cookie soo, keep a hashmap of IDs and user
// data??
//
// check authentication first of everything as it changes decisions on everything and give it to fn
// too
// fn(req, Option<AuthData>, (adrr, etc)) -> Action
//
// and also add an action to... hook in the response without even letting backend know (redirect
// ppl out and such), also in no case we need multiple actions. NOPE, POST login should capture
// body and hijack the response, I'll also just make this flexible too.

#[allow(clippy::type_complexity)]
#[derive(Debug)]
enum ProxyAction<'a> {
    // combining like this eases a lot the implementation
    CaptureBodyAndHijack(String),
    // (
    //     fn(&httparse::Request, &[u8]) -> Result<(), ClientError>,
    //     fn(&httparse::Request, &mut TcpStream) -> Result<(), ClientError>,
    // ),
    // ),
    CaptureResponse(String),
    //fn(&httparse::Request, &[u8]) -> Result<(), ClientError>),
    ContWithMaybeAuth(Option<Ref<'a, String, AuthData>>),
}

impl ProxyAction<'_> {
    // username - password
    #[allow(dead_code)]
    type CaptureBodyConclussion = (String, String);

    pub fn capture_body(
        _req: &httparse::Request,
        _id: &str,
        _body: &[u8],
    ) -> Result<Self::CaptureBodyConclussion, ClientError> {
        todo!();
    }

    pub fn hijack(
        _req: &httparse::Request,
        _body_conclussion: Self::CaptureBodyConclussion,
        _stream: &mut TcpStream,
    ) -> Result<(), ClientError> {
        todo!();
    }

    pub fn capture_response(
        _req: &httparse::Request,
        _res: &httparse::Response,
        id: String,
        csrf_keyname: &str,
        body: &[u8],
    ) -> Result<(), ClientError> {
        // Find csrf token and add it to list of things
        let document = scraper::Html::parse_document(
            str::from_utf8(body).map_err(ClientError::TextEncodingError)?,
        );
        let selector = scraper::Selector::parse(format!("input[name=\"{csrf_keyname}\"]").as_str())
            .map_err(|err| ClientError::SelectorBuildError(format!("{err}")))?;
        let csrf = document
            .select(&selector)
            .next()
            .and_then(|elem| elem.attr("value"))
            .ok_or(ClientError::CsrfNotFound)?;

        println!("inserting {id} - {csrf}");
        CSRF_MAP.insert(id, csrf.to_string());
        Ok(())
    }
}

#[derive(Debug)]
pub struct AuthData {
    pub username: String,
    pub email: String,
    pub fullname: Option<String>,
}
impl AuthData {
    pub fn write_as_headers(&self, into: &mut impl Write) -> Result<(), ClientError> {
        use constants::HTTP_HEADERS_NEWLINE as NL;

        let r: Result<_, io::Error> = try {
            into.write_all(b"X-WEBAUTH-USER: ")?;
            into.write_all(self.username.as_bytes())?;
            into.write_all(NL)?;
            into.write_all(b"X-WEBAUTH-EMAIL: ")?;
            into.write_all(self.email.as_bytes())?;
            into.write_all(NL)?;
            if let Some(fullname) = self.fullname.as_ref() {
                into.write_all(b"X-WEBAUTH-FULLNAME: ")?;
                into.write_all(fullname.as_bytes())?;
                into.write_all(NL)?;
            }
        };
        r.map_err(ClientError::AuthHeadersIoError)
    }
}

fn what_to_do<'a>(
    req: &httparse::Request,
    id_auth: Option<(String, Option<Ref<'a, String, AuthData>>)>,
) -> ProxyAction<'a> {
    let trimedpath = req
        .path
        .and_then(|path| path.split_once('?').map(|spl| spl.0).or(Some(path)));
    // FIXME: auth might not be set up until gitea replies and sets an ID
    match (trimedpath, req.method, id_auth) {
        (Some("/user/login"), Some("GET"), Some((id, None))) => {
            // We gotta get the gitea response to find the csrf token and save that
            ProxyAction::CaptureResponse(id)
        }
        (Some("/user/login"), Some("POST"), Some((id, None))) => {
            // Here we gotta get the post body, check csrf, and check passwd
            // Then get their account data, add it to their ID and redirect back to home
            ProxyAction::CaptureBodyAndHijack(id)
        }
        (_, _, id_auth) => ProxyAction::ContWithMaybeAuth(id_auth.and_then(|id_auth| id_auth.1)),
    }
}

// TODO: might wanna save this one between runtimes
pub static AUTH_MAP: std::sync::LazyLock<DashMap<String, AuthData>> =
    std::sync::LazyLock::new(DashMap::new);
// TODO: time expire old keys?
pub static CSRF_MAP: std::sync::LazyLock<DashMap<String, String>> =
    std::sync::LazyLock::new(DashMap::new);

#[allow(clippy::too_many_lines)]
fn handle_client(client: &mut TcpStream, config: &config::Schema) -> Result<(), ClientError> {
    use ClientError as E;
    use ProxyAction as A;

    http_bodies::parse_http!(client_req, client, Request, let (req, _, body_over));

    let theres_body = req.headers.expect_body();

    let host_header = req.headers.header("host").ok_or(E::HostHeaderNotFound)?;
    let cookies = req.headers.header("cookie");

    // Now find that header and pass everything
    let read_hosts = config.hosts.read();
    let (addr, id_keyname, csrf_keyname) = read_hosts
        .get(host_header.as_ref())
        .ok_or_else(|| E::HostNotRegistered(host_header.to_string()))?
        .clone();
    drop(read_hosts);

    // Try to find auth data
    let id_auth = if let Some(cookies) = cookies {
        if let Some(id_ck) = cookie::Cookie::split_parse(cookies)
            .filter_map(Result::ok)
            .find(|ck| ck.name() == id_keyname)
        {
            let id = id_ck.value().to_string();
            let val = AUTH_MAP.get(&id);
            Some((id, val))
        } else {
            None
        }
    } else {
        None
    };

    let action = what_to_do(&req, id_auth);
    if let A::CaptureBodyAndHijack(id) = action {
        let mut body = body_over.to_vec();
        if theres_body {
            // not sure if this will append or overwrite, wanna append
            io::copy(client, &mut body).map_err(|err| E::BufferError("body", err))?;
        }
        let c = A::capture_body(&req, id.as_ref(), &body[..])?;
        A::hijack(&req, c, client)?;
    } else {
        let mut stream = TcpStream::connect(addr).map_err(E::BackendConnectFail)?;
        // stream.write_all(body_head).map_err(E::ExchangeIoError)?;
        stream
            .write_all(
                format!(
                    "{} {} HTTP/1.1\r\n",
                    req.method.unwrap_or("GET"),
                    req.path.unwrap_or("/")
                )
                .as_bytes(),
            )
            .map_err(E::ExchangeIoError)?;
        req.headers
            .iter()
            .filter(|header| !header.name.eq_ignore_ascii_case("accept-encoding"))
            .map(|header| {
                stream.write_all(
                    &{
                        let mut all = Vec::new();
                        all.extend_from_slice(header.name.as_bytes());
                        all.extend_from_slice(b": ");
                        all.extend_from_slice(header.value);
                        all.extend_from_slice(b"\r\n");
                        all
                    }[..],
                )
            })
            .try_fold((), |(), header| header)
            .map_err(E::ExchangeIoError)?;
        if let A::ContWithMaybeAuth(Some(ref auth)) = action {
            auth.write_as_headers(&mut stream)?;
        }
        stream
            .write_all(constants::HTTP_HEADERS_NEWLINE)
            .map_err(E::ExchangeIoError)?;

        // send our overhead
        stream.write_all(body_over).map_err(E::ExchangeIoError)?;
        if theres_body {
            io::copy(client, &mut stream).map_err(E::ExchangeIoError)?;
        }
        if let A::CaptureResponse(id) = action {
            http_bodies::parse_http!(stream_res, stream, Response, let (resp, _, res_body_over));

            let buf = http_bodies::capture_body(
                resp.headers,
                res_body_over,
                &mut stream,
                "response buffer",
            )?
            .unwrap_or_else(Vec::new);
            A::capture_response(&req, &resp, id, csrf_keyname.as_ref(), &buf[..])?;

            client
                .write_all(
                    format!(
                        "HTTP/1.1 {} {}\r\n",
                        resp.code.unwrap_or(200),
                        resp.reason.unwrap_or("OK")
                    )
                    .as_bytes(),
                )
                .map_err(E::ExchangeIoError)?;
            resp.headers
                .iter()
                .filter(|header| !header.name.eq_ignore_ascii_case("transfer-encoding"))
                .map(|header| {
                    client.write_all(
                        &{
                            let mut all = Vec::new();
                            all.extend_from_slice(header.name.as_bytes());
                            all.extend_from_slice(b": ");
                            all.extend_from_slice(header.value);
                            all.extend_from_slice(b"\r\n");
                            all
                        }[..],
                    )
                })
                .try_fold((), |(), header| header)
                .map_err(E::ExchangeIoError)?;

            // FIXME: :grimacing: (append to headers or smth)
            client
                .write_all(format!("Content-Length: {}\r\n\r\n", buf.len()).as_bytes())
                .map_err(E::ExchangeIoError)?;
            client
                .write_all(constants::HTTP_HEADERS_NEWLINE)
                .map_err(E::ExchangeIoError)?;

            client.write_all(&buf[..]).map_err(E::ExchangeIoError)?;
        } else {
            io::copy(&mut stream, client).map_err(E::ExchangeIoError)?;
        }
    }

    Ok(())
}

mod http_bodies {
    use crate::{listener::ClientError, utils::headers::HeadersExt};

    // pub trait HttpMessage<'h, 'b> {
    //     fn new(headers: &'h mut [httparse::Header<'b>]) -> Self;
    //     fn parse(&mut self, buf: &'b [u8]) -> httparse::Result<usize>;
    // }

    // impl<'h, 'b> HttpMessage<'h, 'b> for httparse::Request<'h, 'b> {
    //     fn new(headers: &'h mut [httparse::Header<'b>]) -> Self {
    //         httparse::Request::new(headers)
    //     }

    //     fn parse(&mut self, buf: &'b [u8]) -> httparse::Result<usize> {
    //         self.parse(buf)
    //     }
    // }

    // impl<'h, 'b> HttpMessage<'h, 'b> for httparse::Response<'h, 'b> {
    //     fn new(headers: &'h mut [httparse::Header<'b>]) -> Self {
    //         httparse::Response::new(headers)
    //     }

    //     fn parse(&mut self, buf: &'b [u8]) -> httparse::Result<usize> {
    //         self.parse(buf)
    //     }
    // }

    pub struct ChunckedTransferEncoding<'a, R: Read> {
        rem: Vec<u8>,
        reader: &'a mut R,
    }
    // This was a read impl, then tried to go into an interator and ended up being... this
    impl<R: Read> ChunckedTransferEncoding<'_, R> {
        fn read(&mut self, mut cb: impl FnMut(&[u8])) -> io::Result<()> {
            loop {
                let Some(busize_after_pos) =
                    self.rem.windows(b"\r\n".len()).position(|w| w == b"\r\n")
                else {
                    let mut tmp = [0u8; 4096];
                    let n = self.reader.read(&mut tmp)?;
                    self.rem.extend_from_slice(&tmp[..n]);
                    continue;
                };

                let chunk_len = self.rem[0..busize_after_pos]
                    .iter()
                    .try_fold(0usize, |mut acc, i| {
                        acc <<= 4;

                        #[allow(clippy::manual_is_ascii_check)]
                        {
                            acc += (if (b'0'..=b'9').contains(i) {
                                i - b'0'
                            } else if (b'A'..=b'F').contains(i) {
                                i - b'A'
                            } else if (b'a'..=b'f').contains(i) {
                                i - b'f'
                            } else {
                                return None;
                            }) as usize;
                        }
                        Some(acc)
                    })
                    .ok_or(io::Error::new(
                        io::ErrorKind::InvalidData,
                        "invalid chunk size head contents",
                    ))?;

                if chunk_len == 0 {
                    return Ok(());
                }

                while self.rem.len() < busize_after_pos + chunk_len + (b"\r\n".len() * 2) {
                    let mut tmp = [0u8; 4096];
                    let n = self.reader.read(&mut tmp)?;
                    self.rem.extend_from_slice(&tmp[..n]);
                }
                cb(&self.rem[(busize_after_pos + b"\r\n".len())
                    ..(busize_after_pos + b"\r\n".len() + chunk_len)]);
                self.rem = self.rem
                    [(busize_after_pos + b"\r\n".len() + chunk_len + b"\r\n".len())..]
                    .to_vec();
            }
        }
    }

    pub fn capture_body(
        headers: &[httparse::Header<'_>],
        body_over: &[u8],
        reader: &mut impl Read,
        debug_bufname: &'static str,
    ) -> Result<Option<Vec<u8>>, ClientError> {
        Ok(if let Some(tenc) = headers.header("transfer-encoding") {
            if tenc != "chunked" {
                return Err(ClientError::InvalidHeader("transfer-encoding"));
            }

            let mut buf = Vec::new();
            (ChunckedTransferEncoding {
                rem: body_over.to_vec(),
                reader,
            })
            .read(|chunk| {
                buf.extend_from_slice(chunk);
            })
            .map_err(|err| ClientError::BufferError(debug_bufname, err))?;
            Some(buf)
        } else if let Some(clen) = headers.header("content-length") {
            let n: usize = clen
                .parse()
                .map_err(|_| ClientError::InvalidHeader("content-length"))?;
            let mut buf = Vec::with_capacity(n);
            buf.extend_from_slice(body_over);
            reader
                .read_exact(&mut buf[body_over.len()..])
                .map_err(|err| ClientError::BufferError(debug_bufname, err))?;
            Some(buf)
        } else {
            None
        })
    }

    // I hate it too
    macro_rules! parse_http {
        ($trash_state:tt, $stream:expr, $msg_type:ty, $ret:stmt) => {
            $crate::listener::http_bodies::parse_http!($trash_state, $stream, $msg_type, 16, {1024 * 8}, $ret);
        };
        ($trash_state:tt, $stream:expr, $msg_type:ty, $headers_len:expr, $buf_size:expr, $ret:stmt) => {
            paste::paste! {
                let mut [<$trash_state _buf>] = [0u8; $buf_size];
                let mut [<$trash_state _pos>] = 0;
                let mut [<$trash_state _headers>];

                $ret = loop {
                    [<$trash_state _headers>] = [httparse::EMPTY_HEADER; $headers_len];
                    let mut res = <$msg_type>::new(&mut [<$trash_state _headers>]);

                    let n = $stream
                        .read(&mut [<$trash_state _buf>][[<$trash_state _pos>]..])
                        .map_err(ClientError::HeadReadError)?;
                    [<$trash_state _pos>] += n;

                    if let Ok(httparse::Status::Complete(n)) = res.parse(&[<$trash_state _buf>][0..[<$trash_state _pos>]]) {
                        break (res, &[<$trash_state _buf>][0..(n - $crate::constants::HTTP_HEADERS_NEWLINE.len())], &[<$trash_state _buf>][n..[<$trash_state _pos>]]);
                    }
                };
            }
        };
    }
    use std::io::{self, Read};

    pub(crate) use parse_http;
}
